package org.gel.mauve;

import java.awt.*;

/**
 * An extension of {@link javax.swing.JList} that supports drag and drop
 * to rearrange its contents and to move objects in and out of the list.
 * The objects in the list will be passed either as a String by calling
 * the object's <tt>toString()</tt> object, or if your drag and drop target
 * accepts the {@link TransferableObject#DATA_FLAVOR} data flavor then
 * the actual object will be passed.
 *
 * <p>I'm releasing this code into the Public Domain. Enjoy.
 * </p>
 * <p><em>Original author: Robert Harder, rharder@usa.net</em></p>
 * <p><em>Adapted for RearrangementPanel by Aaron Darling</em></p>
 *
 * @author  Aaron Darling
 * @version 1.1
 */
public class DnDRearrangementPanel 
extends RearrangementPanel
implements  java.awt.dnd.DropTargetListener, 
            java.awt.dnd.DragSourceListener, 
            java.awt.dnd.DragGestureListener    
{

    private java.awt.dnd.DropTarget dropTarget = null;

    private java.awt.dnd.DragSource dragSource = null;
    
    private int sourceIndex = -1;
    private int dropIndex   = -1;
    private Object sourceObject;

    
    /**
     * Constructs a default {@link DnDRearrangementPanel} using a
     * {@link javax.swing.DefaultListModel}.
     *
     * @since 1.1
     */
    public DnDRearrangementPanel( MauveFrame mauve_frame ) 
    {   
    	super( mauve_frame );
        initComponents();
    }   // end constructor
    
    

    
    private void initComponents()
    {
        dropTarget = new java.awt.dnd.DropTarget (this, this);
        dragSource = new java.awt.dnd.DragSource();
        dragSource.createDefaultDragGestureRecognizer( this, java.awt.dnd.DnDConstants.ACTION_MOVE, this);
    }   // end initComponents
    
    
/* ********  D R A G   G E S T U R E   L I S T E N E R   M E T H O D S  ******** */    
    
    
    
  
    public void dragGestureRecognized( java.awt.dnd.DragGestureEvent event) 
    {   //System.out.println( "DragGestureListener.dragGestureRecognized" );
        java.awt.Point origin = event.getDragOrigin();
        for( int seqI = 0; seqI < sequenceCount; seqI++ ){
        	// determine which sequence panel is getting moved
        	
        	if( ((Component)sequencePanels.elementAt( seqI )).getBounds().contains( origin ) ){
        		// this is the component to d'n'd
				final Integer seqpanelI = new Integer( seqI );
	            java.awt.datatransfer.Transferable transfer = new TransferableObject( seqpanelI.getClass(), new TransferableObject.Fetcher()
	            {   /**
	                 * This will be called when the transfer data is requested at the very end.
	                 * At this point we can remove the object from its original place in the list.
	                 */
	                public Object getObject()
	                {   
	                    return seqpanelI;
	                }   // end getObject
	            }); // end fetcher
	            
	            // as the name suggests, starts the dragging
				if( MauveFrame.fist_cursor != null )
		            dragSource.startDrag ( event, MauveFrame.fist_cursor, transfer, this);
	            else
		            dragSource.startDrag (event, java.awt.dnd.DragSource.DefaultLinkDrop, transfer, this);
        		
        	}
        }
    }   // end dragGestureRecognized

    
/* ********  D R A G   S O U R C E   L I S T E N E R   M E T H O D S  ******** */    
    
    
    public void dragDropEnd( java.awt.dnd.DragSourceDropEvent evt ) 
    {   //System.out.println( "DragSourceListener.dragDropEnd" );
    }   // end dragDropEnd

    
    public void dragEnter( java.awt.dnd.DragSourceDragEvent evt )
    {   //System.out.println( "DragSourceListener.dragEnter" );
    }   // end dragEnter
    
    
    public void dragExit( java.awt.dnd.DragSourceEvent evt )
    {   //System.out.println( "DragSourceListener.dragExit" );
    }   // end dragExit
    
    
    public void dragOver( java.awt.dnd.DragSourceDragEvent evt )
    {   //System.out.println( "DragSourceListener.dragOver" );
    }   // end dragOver
    
    
    public void dropActionChanged( java.awt.dnd.DragSourceDragEvent evt )
    {   //System.out.println( "DragSourceListener.dropActionChanged" );
    }   // end dropActionChanged
   
    
    
/* ********  D R O P   T A R G E T   L I S T E N E R   M E T H O D S  ******** */    
    
    
    
    public void dragEnter( java.awt.dnd.DropTargetDragEvent evt ) 
    {   //System.out.println( "DropTargetListener.dragEnter" );
        evt.acceptDrag( java.awt.dnd.DnDConstants.ACTION_MOVE);
    }   // end dragEnter

    public void dragExit( java.awt.dnd.DropTargetEvent evt )
    {   //System.out.println( "DropTargetListener.dragExit" );
    }   // end dragExit
    
    public void dragOver( java.awt.dnd.DropTargetDragEvent evt )
    {   //System.out.println( "DropTargetListener.dragOver" );
    }   // end dragOver
    
    public void dropActionChanged( java.awt.dnd.DropTargetDragEvent evt )
    {   //System.out.println( "DropTargetListener.dropActionChanged" );
        evt.acceptDrag( java.awt.dnd.DnDConstants.ACTION_MOVE);
    }   // end dropActionChanged
    
    
    
    public void drop( java.awt.dnd.DropTargetDropEvent evt ) 
    {   //System.out.println( "DropTargetListener.drop" );
        java.awt.datatransfer.Transferable transferable = evt.getTransferable();
        
        // If it's our native TransferableObject, use that
        if( transferable.isDataFlavorSupported( java.awt.datatransfer.DataFlavor.javaFileListFlavor ) ||
        	transferable.isDataFlavorSupported( java.awt.datatransfer.DataFlavor.stringFlavor  ) )
        {
            evt.acceptDrop( java.awt.dnd.DnDConstants.ACTION_MOVE );
            Object obj = null;
	        final Integer seqpanelJ = new Integer(0);
            Class classy = seqpanelJ.getClass();
            try
            {
                obj = transferable.getTransferData( new java.awt.datatransfer.DataFlavor( classy , TransferableObject.MIME_TYPE ) );
            }   // end try
            catch( java.awt.datatransfer.UnsupportedFlavorException e )
            {   e.printStackTrace();
            }   // end catch
            catch( java.io.IOException e )
            {   e.printStackTrace();
            }   // end catch

            if( obj != null )
            {
				// See where in the list we dropped the element.
		        java.awt.Point origin = evt.getLocation();
				sourceIndex = ((Integer)obj).intValue();
				dropIndex = 0;
		        for( int seqI = 0; seqI < sequenceCount; seqI++ ){
		        	// determine which panel the drop occurred over
		        	if( ((Component)sequencePanels.elementAt( seqI )).getBounds().contains( origin ) ||
		        		((Component)rulerPanels.elementAt( seqI )).getBounds().contains( origin ) ){
		        		dropIndex = seqI;
		        		break;
		        	}
		        }

				if( dropIndex != sourceIndex ){
					int[] new_order = new int[ sequenceCount ];
					int seqI = 0, seqJ = 0;
					for( ; seqI < sequenceCount; ){
						if( seqI == dropIndex ){
							new_order[ seqI ] = sourceIndex;
							seqI++;
						}else if( seqJ == sourceIndex ){
							// skip this one
							seqJ++;
						}else{
							new_order[ seqI ] = seqJ;
							seqI++;
							seqJ++;
						}
					}
					reorderSequences( new_order );
				}

				evt.getDropTargetContext().dropComplete(true);
            }   // end if: we got the object
            
            // Else there was a problem getting the object
            else
            {
                evt.rejectDrop();
            }   // end else: can't get the object
        }   // end if: it's a native TransferableObject
        
        // Else we can't handle this
        else evt.rejectDrop();
    }   // end drop


  
}   // end class DnDRearrangementPanel
