package org.gel.mauve;

import java.util.*;

/**
 * The LCB class tracks locally collinear blocks:  regions of homologous sequence that
 * do not contain rearrangements.
 */
class LCB{
	
	public LCB(){}
	
	public LCB( LCB l ){
		int seq_count = l.left_end.length;
		left_end = new long[ seq_count ];
		right_end = new long[ seq_count ];
		left_adjacency = new int[ seq_count ];
		right_adjacency = new int[ seq_count ];
		reverse = new boolean[ seq_count ];
		lcb_id = l.lcb_id;
		weight = l.weight;
		color = l.color;
		match_color = l.match_color;
		keep = l.keep;
		
		System.arraycopy( l.left_end, 0, left_end, 0, seq_count );
		System.arraycopy( l.right_end, 0, right_end, 0, seq_count );
		System.arraycopy( l.left_adjacency, 0, left_adjacency, 0, seq_count );
		System.arraycopy( l.right_adjacency, 0, right_adjacency, 0, seq_count );
		System.arraycopy( l.reverse, 0, reverse, 0, seq_count );
	}

	/**
	 * Reorder the sequences to conform to the order given in new_order[]
	 */
	public void reorder( int new_order[] ){
		long[] tmp = null;
		int[] tmp_int = null;
		int seqI = 0;

		// reorder left_end
		tmp = new long[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			tmp[ seqI ] = left_end[ new_order[ seqI ] ];
		left_end = tmp;

		// reorder right_end
		tmp = new long[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			tmp[ seqI ] = right_end[ new_order[ seqI ] ];
		right_end = tmp;

		// reorder left_adjacency
		tmp_int = new int[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			tmp_int[ seqI ] = left_adjacency[ new_order[ seqI ] ];
		left_adjacency = tmp_int;

		// reorder right_adjacency
		tmp_int = new int[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			tmp_int[ seqI ] = right_adjacency[ new_order[ seqI ] ];
		right_adjacency = tmp_int;

		// reorder reverse
		boolean[] rev_tmp = new boolean[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			rev_tmp[ seqI ] = reverse[ new_order[ seqI ] ];
		reverse = rev_tmp;
		
		// correct the orientation so the first sequence becomes reference orientation
		if( reverse[ 0 ] )
			for( seqI = 0; seqI < new_order.length; seqI++ )
				reverse[ seqI ] = !reverse[ seqI ];
	}
	
	public long[] left_end;	/**< The left end position of the LCB in each sequence */
	public long[] right_end;  /**< The right end position of the LCB in each sequence */
	public int[] left_adjacency;	/**< 'Pointers' (actually IDs) to the LCBs on the left in each sequence */
	public int[] right_adjacency;	/**< 'Pointers' (actually IDs) to the LCBs on the right in each sequence */
	int lcb_id;			/**< A numerical ID that can be assigned to this LCB */
	long weight;		/**< The weight (or coverage) of this LCB */
	public boolean[] reverse;	/**< The orientation of the LCB in each sequence */
	java.awt.Color color;		/**< The color of the LCB frame */
	java.awt.Color match_color;	/**< The color of matches within the LCB */
	
	boolean keep;	/**< set this to true to keep this LCB even when it's weight is too low */
}

/**
 * Compares left end of LCBs.
 */
class LCBLeftComparator implements Comparator {
	LCBLeftComparator( int seq ){
		m_seq = seq;
	}
	LCBLeftComparator( LCBLeftComparator lmc ){
		m_seq = lmc.m_seq;
	}

	protected int m_seq;

	public int compare( Object o_a, Object o_b) {
		
		LCB a = (LCB)o_a;
		LCB b = (LCB)o_b;
		
		long a_start = a.left_end[ m_seq ], b_start = b.left_end[ m_seq ];
		if( a_start == 0 || b_start == 0 ){
			if( b_start != 0 )
				return 0;
			return 0;
		}

		long diff = a_start - b_start;
		return (int)diff;
	}

};


class LCBlist {
	
	/**
	 *  Redesign to be more intuitive.  left_adjacency is always left, regardless of LCB orientation
	 */
	static void computeLCBAdjacencies( LCB[] lcb_list ){
		if( lcb_list.length == 0 )
			return;	// there aren't any LCBs so there aren't any adjacencies!

		int seq_count = lcb_list[ 0 ].left_end.length;
		LCBLeftComparator llc = new LCBLeftComparator( 0 );
		Arrays.sort( lcb_list, llc );

		int seqI;
		int lcbI;
		for( lcbI = 0; lcbI < lcb_list.length; lcbI++ ){
			lcb_list[ lcbI ].left_adjacency = new int[ seq_count ];
			lcb_list[ lcbI ].right_adjacency = new int[ seq_count ];
			lcb_list[ lcbI ].lcb_id = lcbI;
		}

		for( seqI = 0; seqI < seq_count; seqI++ ){
			llc = new LCBLeftComparator( seqI );
			Arrays.sort( lcb_list, llc );

			for( lcbI = 1; lcbI + 1 < lcb_list.length; lcbI++ ){
				lcb_list[ lcbI ].left_adjacency[ seqI ] = lcb_list[ lcbI - 1 ].lcb_id;
				lcb_list[ lcbI ].right_adjacency[ seqI ] = lcb_list[ lcbI + 1 ].lcb_id;
			}
			if( lcbI == lcb_list.length )
				lcbI--;	// need to decrement when there is only a single LCB

			// set first and last lcb adjacencies to -1
			lcb_list[ 0 ].left_adjacency[ seqI ] = -1;
			lcb_list[ lcbI ].right_adjacency[ seqI ] = -1;
			if( lcbI > 0 ){
				lcb_list[ 0 ].right_adjacency[ seqI ] = lcb_list[ 1 ].lcb_id;
				lcb_list[ lcbI ].left_adjacency[ seqI ] = lcb_list[ lcbI - 1 ].lcb_id;
			}
		}

		// sort back on the first genome order
		llc = new LCBLeftComparator( 0 );
		Arrays.sort( lcb_list, llc );
	}

	
	public static void greedyBreakpointElimination( long minimum_weight, LCB[] adjacencies, Vector change_points ){
		if( adjacencies.length == 0 )
			return;
		int sequenceCount = adjacencies[ 0 ].left_end.length;
		
		// repeatedly remove the low weight LCBs until the minimum weight criteria is satisfied
		int lcbI = 0;
		long min_weight = 0;
		long prev_min_weight = 0;
		int min_lcb = 0;
		int lcb_count = adjacencies.length;
		boolean status_out = true;

		while( min_weight < minimum_weight ){

			min_weight = 0;
			for( lcbI = 0; lcbI < adjacencies.length; lcbI++ ){
				if( adjacencies[ lcbI ].lcb_id != lcbI ){
					// this lcb has been removed or merged with another lcb
					continue;
				}
				if( adjacencies[ lcbI ].keep == true )
					// The user has requested this LCB never be removed
					continue;

				if( adjacencies[ lcbI ].weight < min_weight || min_weight == 0 ){
					min_weight = adjacencies[ lcbI ].weight;
					min_lcb = lcbI;
				}
			}
			lcbI = min_lcb;

			if( prev_min_weight != min_weight ){
				if( change_points != null )
					change_points.addElement( new Integer( (int)min_weight ) );
				
//				if( status_out != false )
//					System.out.print( "There are " + lcb_count + " LCBs with minimum weight " + min_weight + "\n" );;
			}
						
			if( lcb_count == 1 )
				break;	// if only a single LCB remains, don't remove it

			if( min_weight >= minimum_weight )
				break;

			prev_min_weight = min_weight;

			// actually remove the LCBs now
			// (only remove a single LCB for now -- it's easier to calculate adjacencies)

			// remove this LCB
			adjacencies[ lcbI ].lcb_id = -2;
			
			// update adjacencies
			int seqI;
			int left_adj;
			int right_adj;
			for( seqI = 0; seqI < sequenceCount; seqI++ ){
				left_adj = adjacencies[ lcbI ].left_adjacency[ seqI ];
				right_adj = adjacencies[ lcbI ].right_adjacency[ seqI ];

				if( left_adj == -2 || right_adj == -2 ){
					System.err.print( "improper linking\n" );
				}
				// for debugging, check for consistency:
				if( left_adj != -1 && adjacencies[ left_adj ].right_adjacency[ seqI ] != lcbI )
					System.err.print( "Mutiny on the bounty!\n" );
				// for debugging, check for consistency
				if( right_adj == adjacencies.length )
					System.err.print( "Horrible Error -399a\n" );
				if( right_adj != -1 && adjacencies[ right_adj ].left_adjacency[ seqI ] != lcbI )
					System.err.print( "Mutiny on the bounty!\n" );
				if( left_adj != -1 )
					adjacencies[ left_adj ].right_adjacency[ seqI ] = right_adj;
				if( right_adj != -1 && right_adj != adjacencies.length )
					adjacencies[ right_adj ].left_adjacency[ seqI ] = left_adj;
				
			}
			// just deleted an lcb, drop the lcb count
			lcb_count--;

			// check for collapse
			for( seqI = 0; seqI < sequenceCount; seqI++ ){
				left_adj = adjacencies[ lcbI ].left_adjacency[ seqI ];
				right_adj = adjacencies[ lcbI ].right_adjacency[ seqI ];
				if( right_adj == adjacencies.length )
					System.err.print( "Horrible Error -399a\n" );
				if( left_adj == -1 || right_adj == -1 )
					continue;	// can't collapse with a non-existant LCB!
				// check whether this LCB has already been merged
				if( left_adj != adjacencies[ left_adj ].lcb_id ||
					right_adj != adjacencies[ right_adj ].lcb_id ){
					// because adjacency pointers are always updated to point to the 
					// representative entry of an LCB, the lcb_id and the array index
					// should always be identical
					System.err.print( "improper linking\n" );
					continue;
				}
				if( left_adj == -2 || right_adj == -2 ){
					System.err.print( "improper linking\n" );
				}
				// check whether the two LCBs are adjacent in each sequence
				boolean orientation = !adjacencies[ left_adj ].reverse[ seqI ];
				int seqJ;
				for( seqJ = 0; seqJ < sequenceCount; seqJ++ ){
					boolean j_orientation = !adjacencies[ left_adj ].reverse[ seqJ ];
					if( j_orientation == orientation &&
						adjacencies[ left_adj ].right_adjacency[ seqJ ] != right_adj )
						break;
					if( j_orientation != orientation &&
						adjacencies[ left_adj ].left_adjacency[ seqJ ] != right_adj )
						break;
					// check that they are both in the same orientation
					if( (!adjacencies[ right_adj ].reverse[ seqJ ]) != j_orientation )
						break;
				}

				if( seqJ != sequenceCount )
					continue;
				
				// these two can be collapsed.  repeatedly search the intervening region

					
				adjacencies[ right_adj ].lcb_id = left_adj;
				if( adjacencies[ right_adj ].lcb_id == -1 ||
					adjacencies[ right_adj ].lcb_id == -2 )
					System.err.print( "Trouble in the eleventh circle\n" );
				adjacencies[ left_adj ].weight += adjacencies[ right_adj ].weight;
				// unlink right_adj from the adjacency list and
				// update left and right ends of left_adj
				for( seqJ = 0; seqJ < sequenceCount; seqJ++ ){
					boolean j_orientation = !adjacencies[ left_adj ].reverse[ seqJ ];
					int rr_adj = adjacencies[ right_adj ].right_adjacency[ seqJ ];
					int rl_adj = adjacencies[ right_adj ].left_adjacency[ seqJ ];
					if( j_orientation == orientation ){
						adjacencies[ left_adj ].right_end[ seqJ ] = adjacencies[ right_adj ].right_end[ seqJ ];
						adjacencies[ left_adj ].right_adjacency[ seqJ ] = rr_adj;
						if( rr_adj == adjacencies.length )
							System.err.print( "Horrible Error -399a\n" );
						if( rr_adj != -1 )
							adjacencies[ rr_adj ].left_adjacency[ seqJ ] = left_adj;
					}else{
						adjacencies[ left_adj ].left_end[ seqJ ] = adjacencies[ right_adj ].left_end[ seqJ ];
						adjacencies[ left_adj ].left_adjacency[ seqJ ] = rl_adj;
						if( rl_adj == adjacencies.length )
							System.err.print( "Horrible Error -399a\n" );
						if( rl_adj != -1 )
							adjacencies[ rl_adj ].right_adjacency[ seqJ ] = left_adj;
					}
					// update lcbI's adjacency links to point nowhere
					if( adjacencies[ lcbI ].left_adjacency[ seqJ ] == right_adj )
						adjacencies[ lcbI ].left_adjacency[ seqJ ] = left_adj;
					if( adjacencies[ lcbI ].right_adjacency[ seqJ ] == right_adj )
						adjacencies[ lcbI ].right_adjacency[ seqJ ] = left_adj;
				}

				// just collapsed an lcb, decrement lcb_count
				lcb_count--;
			}
		}

//		if( status_out != false )
//			System.out.print( "There are " + lcb_count + " LCBs with minimum weight " + min_weight + "\n" );;
		if( change_points != null )
			change_points.addElement( new Integer( (int)min_weight ) );

	}
	/**
	 * Note: this code assumes that matches and LCBs are sorted on their coordinates
	 * in the first genome sequence
	 */
	public static LCB[] filterLCBs( LCB[] adjacencies, Vector matches, Vector removed_lcbs ){
		if( adjacencies.length == 0 )
			return adjacencies;

		int lcbI;
		for( lcbI = 0; lcbI < adjacencies.length; lcbI++ ){
			if( adjacencies[ lcbI ].lcb_id == lcbI ){
				continue;
			}
			if( adjacencies[ lcbI ].lcb_id == -1 ){
				System.err.println( "weird" );	// don't want any weirdness
				continue; 	// something's wrong
			}
			if( adjacencies[ lcbI ].lcb_id == -2 )
				continue; 	// this one was removed

			// this one points elsewhere
			// search and update the union/find structure for the target
			Stack visited_lcbs = new Stack();
			visited_lcbs.push( new Integer( lcbI ) );
			int cur_lcb = adjacencies[ lcbI ].lcb_id;
			while( adjacencies[ cur_lcb ].lcb_id != cur_lcb ){
				visited_lcbs.push( new Integer( cur_lcb ) );
				cur_lcb = adjacencies[ cur_lcb ].lcb_id;
				if( cur_lcb == -1 || cur_lcb == -2 ){
					break;
				}
			}
			while( visited_lcbs.size() > 0 ){
				adjacencies[ ((Integer)visited_lcbs.peek()).intValue() ].lcb_id = cur_lcb;
				visited_lcbs.pop();
			}
		}

		// update lcb_ids and colors for the matches
		if( matches != null ){
			for( int matchI = 0; matchI < matches.size(); matchI++ ){
				Match m = (Match)matches.elementAt( matchI );
				m.lcb = adjacencies[ m.lcb ].lcb_id;
				if( m.lcb >= 0 )
					m.color = adjacencies[ m.lcb ].match_color;
			}
		}
		
		// leave only the remaining LCBs
		int remaining_count = 0;
		for( lcbI = 0; lcbI < adjacencies.length; lcbI++ ){
			if( adjacencies[ lcbI ].lcb_id == lcbI )
				remaining_count++;
		}
		LCB[] remainders = new LCB[ remaining_count ];
		remaining_count = 0;
		int match_lcbI = 0;
		for( lcbI = 0; lcbI < adjacencies.length; lcbI++ ){
			if( adjacencies[ lcbI ].lcb_id == lcbI ){
				remainders[ remaining_count ] = adjacencies[ lcbI ];				
				remaining_count++;
			}
		}

		// update the lcb adjacencies!!
		computeLCBAdjacencies( remainders );
		
		// if we're supposed to return the removed LCBs then go get them
		if( removed_lcbs != null ){
			// count the removed LCBs
			int removed_count = 0;
			for( lcbI = 0; lcbI < adjacencies.length; lcbI++ )
				if( adjacencies[ lcbI ].lcb_id == -2 )
					removed_count++;
			// go get the removed LCBs
			LCB[] removed = new LCB[ removed_count ];
			int removedI = 0;
			for( lcbI = 0; lcbI < adjacencies.length; lcbI++ ){
				if( adjacencies[ lcbI ].lcb_id == -2 )
					removed[ removedI++ ] = adjacencies[ lcbI ];
			}
			computeLCBAdjacencies( removed );
			removed_lcbs.add( removed );
		}

		// update the lcb_id inside of each match to point at the new array index
		if( matches != null ){
			for( lcbI = 0; lcbI < remainders.length; lcbI++ ){
				for(; match_lcbI < matches.size(); match_lcbI++ ){
					Match m = (Match)matches.elementAt( match_lcbI );
					if( m.lcb < 0 )
						continue;
					if( m.lcb < lcbI )
						System.err.println( "Some kind of bad error occurred\n" );
					if( m.lcb != lcbI )
						break;
					m.lcb = remainders[ lcbI ].lcb_id;
				}
			}
		}
		
		return remainders;
	}

}