package org.gel.mauve;

import java.awt.event.*;
import java.applet.Applet;
import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import javax.swing.*;

/**
 * A panel with a ruler drawn in it.  This class extends the Panel class to
 * create a panel which contains a ruler of configurable size and tick width.
 */
public class RulerPanel extends JPanel {
    
    long firstTick, lastTick;
    long firstViewable;	/**< The coordinate of the left-most visible tick  */
    long lastViewable;	/**< The coordinate of the right-most visible tick */
	int ticks;		/**< The number of visible ticks */
	int subticks;	/**< The number of subticks between each tick */
	long roundValue; /**< Numbers will be rounded to the nearest multiple of this value */
	int labelPrecision; /**< The precision of the numeric labels */
	
	double top_offset_pct = .2;		/**< Distance from the top of the panel to draw the base line, relative to ruler height */
	double tick_height_pct = .25;	/**< Height (in pixels) of major tick marks relative to ruler height */
	double subtick_height_pct = .12;	/**< Height of minor tick marks relative to ruler height */
	double font_size_pct = .45;		/**< Size of tick label font relative to ruler height */

	final int topOffset = 10;	
	final int tickHeight = 5;	
	final int subtickHeight = 2;	
	final int labelOffset = 22;		
    
    boolean selected = false;

    Color lineColor = new Color(32, 32, 32);	/**< The color of the ruler lines and marks */
    Color textColor = new Color(0, 0, 0);		/**< The color of the ruler numbers */

	/**
	 * Constructs a RulerPanel with the specified coordinate range.
	 * @param first The left-most coordinate that can be displayed on the ruler
	 * @param last The right-most coordinate that can be displayed on the ruler
	 */
    public RulerPanel (long first, long last ) {
//		setLayout(null);
		firstTick = first;
		lastTick = last;
		ticks = 5;
		subticks = 5;
		firstViewable = firstTick;
		lastViewable = lastTick;
		labelPrecision = 1000;
		long precision = (last - first) / labelPrecision;
		roundValue = 1;
		while( roundValue < precision )
			roundValue *= 10;
    }
	
	/**
	 * Constructs a RulerPanel with the specified coordinate range.
	 * @param first The left-most coordinate that can be displayed on the ruler
	 * @param last The right-most coordinate that can be displayed on the ruler
	 * @param tix The number of visible ticks.  Defaults to 5
	 * @param subtix The number of subticks between each pair of ruler ticks.  Defaults to 5
	 */
    public RulerPanel (long first, long last, int tix, int subtix ) {
//		setLayout(null);
		firstTick = first;
		lastTick = last;
		firstViewable = firstTick;
		lastViewable = lastTick;
		ticks = tix;
		subticks = subtix;
		labelPrecision = 1000;
		long precision = (last - first) / labelPrecision;
		roundValue = 1;
		while( roundValue < precision )
			roundValue *= 10;
	}

	/**
	 * Sets the RulerPanel's displayed coordinate range and redraws the ruler.
	 * @param first The left-most coordinate that can be displayed on the ruler
	 * @param length The length of the range of coordinates to display on the ruler
	 */
	public void setViewableRange( long first, long length ) {
		firstViewable = first;
		lastViewable = first + length - 1;
		long precision = length / labelPrecision;
		roundValue = 1;
		while( roundValue < precision )
			roundValue *= 10;
		repaint();
	}
	
	/**
	 * Returns the RRSequencePanel's currently displayed coordinate range in the supplied array.
	 * The parameter @param view_array should be an array of longs with size 2.
	 * The first element will contain the start of the coordinate range and the second element 
	 * will contain the length of the coordinate range.
	 */
	public void getViewableRange( long[] view_array) {
		view_array[0] = firstViewable;
		view_array[1] = lastViewable - firstViewable + 1;
	}
	
	public void setRoundValue( long round_value ){
		roundValue = round_value;
	}
	
	public long getRoundValue(){
		return roundValue;
	}

    public void paint (Graphics graphics) {
		
		Graphics2D g = (Graphics2D)graphics;
		Dimension dim = getSize();
		
		double tickerval = (double)(lastViewable - firstViewable) / (double)ticks;
		double tickerwidth = dim.width / (double)ticks;
		double subtickerwidth = tickerwidth / (double)subticks;
		
		// erase the background
		g.setColor( RearrangementPanel.bg_color );
		g.fillRect( 0, 0, getSize().width, getSize().height );
//		g.clearRect( 0, 0, getSize().width, getSize().height );
		int top_offset = (int)(top_offset_pct * dim.height);
		int tick_height = (int)(tick_height_pct * dim.height);
		int subtick_height = (int)(subtick_height_pct * dim.height);

		Font f = new Font( "Default", Font.PLAIN, (int)(font_size_pct * dim.height) );
		g.setFont( f );

		// draw the base line
		g.setColor( lineColor );
		g.drawLine( 0, top_offset, getSize().width, top_offset );
		
		// draw the ticks and subticks
		int tickBase = 1;
		int subtickX;
		for ( int i = 0; i < ticks; i++ ){
			int tickX = (int)(tickerwidth * i + tickBase);
		    g.drawLine( tickX, top_offset , tickX, top_offset + tick_height);

			for ( int subI = 0; subI < subticks; subI++ ){
				subtickX = (int)(subtickerwidth * subI + tickX);
			    g.drawLine( subtickX, top_offset , subtickX, top_offset + subtick_height);
			}
		}
		
		// draw the tick labels
	    String i_string;
		g.setColor( textColor );
		long[] label_values = new long[ ticks + 1 ];
		String label_suffix = "";
		for ( int i = 0; i <= ticks; i++ ){
			long label_val = (long)(i * tickerval) + firstViewable;
			if( roundValue > 0 ){
				long label_rem = label_val % roundValue;
				if( label_rem < roundValue - label_rem ){
					label_val -= label_rem;
				}else{
					label_val += roundValue - label_rem;
				}
			}
			label_values[ i ] = label_val;
			if( label_val % 1000000000 == 0 && label_suffix == "g" || label_suffix == "" ){
				label_suffix = "g";
			}else
			if( label_val % 1000000 == 0 && label_suffix != "k" && label_suffix != "n" ){
				label_suffix = "m";
			}else
			if( label_val % 1000 == 0 && label_suffix != "n" ){
				label_suffix = "k";
			}else{
				label_suffix = "n";
			}
		}
		for ( int i = 0; i <= ticks; i++ ){
			int tickX = (int)(tickerwidth * i + tickBase) + 2;
			if( label_suffix == "g" )
				label_values[ i ] /= 1000000000;
			if( label_suffix == "m" )
				label_values[ i ] /= 1000000;
			if( label_suffix == "k" )
				label_values[ i ] /= 1000;
			if( label_suffix == "n" )
				label_suffix = "";
			
		    i_string = new Long( label_values[ i ] ).toString() + label_suffix;
		    if( i == ticks ){
				// move the label back
				FontMetrics fm = g.getFontMetrics();
				tickX -= fm.stringWidth( i_string ) + 3;
		    }
		    int font_top = (int)((top_offset_pct + tick_height_pct + font_size_pct) * dim.height );
		    g.drawString(i_string, tickX, font_top ); 
 		}
    }

	public void paintComponent( Graphics g ) {
		paint( g );
	}
}
