package org.gel.mauve;
import java.io.*;
import java.util.*;

/**
 * XMFA file class
 * A technical cross-product of insomnia and indulgence
 */
 
public class XMFAAlignment {
	protected RandomAccessFile xmfa_file;	/**< The file containing alignment data */
	protected Match[] intervals;		/**< The match class instances store the boundaries of each aligned segment */
	protected Match[] file_pos;		/**< These store the offset within each file where an interval's sequence begins */
	protected int newline_size = 0;	/**< The number of newline bytes, as detected on the first line of the file */
	protected int line_width = 80;	/**< The number of chars on each line */
	protected int seq_count = 0;	/**< The number of sequences aligned */
	protected long[] seq_length;	/**< The length of each sequence */
	protected LCB[] lcb_list;		/**< A list of LCBs contained in this alignment */
	protected String[] comments;	/**< Array of comment lines for each alignment entry in the XMFA */
	protected String[] names;		/**< Array of names for each sequence in the first alignment entry */
	
	GappedIntervalSequenceTree[][] gis_tree;	/**< A set of gist's that track map sequence index and column index to file offset. 
													 indexed by [interval][sequence] */

	protected int buffer_size = 500000;	/**< The size of data chunks to read from disk  */

	/**
	 * Constructor reads an XMFA alignment from an input file
	 * and indexes the location of alignment bounds
	 */
	public XMFAAlignment( RandomAccessFile ir ) throws java.io.IOException {
		xmfa_file = ir;
		// simple parse of the input
		Vector lend = new Vector();
		Vector rend = new Vector();
		Vector reverse = new Vector();
		Vector f_offset = new Vector();
		Vector f_end_offset = new Vector();
		Vector tmp_ivs = new Vector();
		Vector tmp_offs = new Vector();
		Vector tmp_lcbs = new Vector();
		Vector gis_tree_tmp = new Vector();
		Vector tmp_comments = new Vector();
		Vector tmp_names = new Vector();
		String cur_comment = null;
		
		boolean gap_state = false;	/**< true when in the midst of reading gaps */
		long block_f_offset = 0;
		long seq_offset = 0;
		long block_seq_offset = 0;
		GappedIntervalSequenceTree gist = null;

		FileKey fk = null;
		GapKey gk = null;
		
		// XMFA file parse states
		final int wait_defline = 0;
		final int read_defline = 1;
		final int read_comment = 2;
		final int read_sequence = 3;
		final int process_sequence = 4;
		
		byte[] buf = new byte[ buffer_size ];
		int bufI = 0;
		int remaining_bytes = 0;
		long cur_base = 0;
		int state = wait_defline;
		int section_start = buffer_size;
		
		// detect the number of bytes in a newline
		xmfa_file.seek( 0 );
		String first_line = xmfa_file.readLine();
		newline_size = (int)(xmfa_file.getFilePointer() - first_line.length());
		xmfa_file.seek( 0 );
		
		// begin parse loop
		while(true){
			if( remaining_bytes == 0 ){
				// it's important to preserve the defline read thus far
				// shift it to the beginning of the array and then read
				// from that point onward
				if( state != read_defline && state != read_comment )
					section_start = buf.length;
				int copy_len = buf.length - section_start;
				System.arraycopy( buf, section_start, buf, 0, copy_len );

				if( state == read_defline || state == read_comment )
					section_start = 0;

				// read more data into the buffer
				cur_base += bufI;
				xmfa_file.seek( cur_base );
				remaining_bytes = xmfa_file.read( buf, copy_len, buf.length - copy_len );
				cur_base -= copy_len;
				if( remaining_bytes <= 0 )
					break;
				bufI = copy_len;	// continue wherever we left off
			}
			
			switch( state ){
			case wait_defline:
				if( buf[ bufI ] == '>' ){
					section_start = bufI;
					state = read_defline;
					bufI--;
					remaining_bytes++;
				}
				break;
			
			// when a newline rolls around, parse out the defline
			case read_defline:
				if( buf[ bufI ] == '\r' || buf[ bufI ] == '\n' ){
					// read the goods into cur_line
					String cur_line = new String( buf, section_start, bufI - section_start );

					// parse a sequence left-end, right-end, and strand combination
					// > name:start-end strand(+/-)
					int colon_pos = cur_line.indexOf( ':' );
					String tmp_str = cur_line.substring( colon_pos+1 );
					int dash_pos = tmp_str.indexOf( '-' );
					int space_pos = tmp_str.indexOf( ' ' );
					long first = Long.parseLong( tmp_str.substring(0,dash_pos) );
					long second = Long.parseLong( tmp_str.substring( dash_pos+1, space_pos ) );
					long left = first < second ? first : second;
					long right = first < second ? second : first;
					boolean cur_reverse = true;
					// parse strand -- if + isn't found then assume - strand 
					if( tmp_str.indexOf( '+', space_pos+1 ) < 0 )
						reverse.addElement( new Boolean( true ) );
					else{
						reverse.addElement( new Boolean( false ) );
						cur_reverse = false;
					}
					lend.addElement( new Long( left ) );
					rend.addElement( new Long( right ) );
					

					// if we haven't yet completed the first alignment entry
					// try to parse the sequence name
					if( tmp_comments.size() == 0 ){
						int strand_pos = 0;
						if( !cur_reverse )
							strand_pos = tmp_str.indexOf( '+', space_pos+1 );
						else
							strand_pos = tmp_str.indexOf( '-', space_pos+1 );
						
						if( tmp_str.length() > strand_pos + 2 )
							// there's a name to parse
							tmp_names.add( tmp_str.substring( strand_pos + 2 ) );
						else
							tmp_names.add( new String("") );
					}

					// prepare for seq parsing
					block_f_offset = -1;
					seq_offset = 0;
					gap_state = false;
					state = read_sequence;
					gist = new GappedIntervalSequenceTree();
					gis_tree_tmp.addElement( gist );
				}
				break;
			
			// ride out the comment
			case read_comment:
				if( buf[ bufI ] == '\r' || buf[ bufI ] == '\n' ){
					// store the comment
					if( cur_comment == null ){
						cur_comment = new String( buf, section_start, bufI - section_start );
						tmp_comments.add( cur_comment );
					}
					if( buf[ bufI ] == '\n' ){
						state = wait_defline;
						cur_comment = null;
					}
				}
				break;

			// wait for a sequence entry to finish, add gist keys, etc.
			case read_sequence:
				if( buf[ bufI ] == '>' || buf[ bufI ] == '=' ){
					state = process_sequence;
					remaining_bytes++;
					bufI--;
					break;
				}

				// do nothing on newlines
				if( buf[ bufI ] == '\r' || buf[ bufI ] == '\n' ){
					break;
				}

				// set the sequence data file offset if this is the beginning of the entry
				if( fk == null && gk == null ){
					f_offset.addElement( new Long( cur_base + bufI ) );
				}

				// it's all sequence data
				long cur_len = gist.length();
				if( buf[ bufI ] == '-' ){
					if( fk != null ){
						gist.insert( fk, GappedIntervalSequenceTree.end );
						if( cur_len + fk.getLength() != gist.length() )
							System.err.println( "crap-o-la!" );
						fk = null;
					}
					if( gk == null )
						gk = new GapKey( 0 );
					gk.length++;
				}else{
					if( gk != null ){
						gist.insert( gk, GappedIntervalSequenceTree.end );
						if( cur_len + gk.getLength() != gist.length() )
							System.err.println( "crap-o-la deux!" );
						gk = null;
					}
					if( fk == null ){
						fk = new FileKey( this, seq_offset, 0, cur_base + bufI, 0 );
					}
					fk.length++;
					fk.f_length = cur_base + bufI - fk.f_offset + 1;
				}

				break;

			// process a completed sequence entry
			case process_sequence:

				// do final processing from a previously parsed sequence
				long curry_len = gist.length();
				if( fk != null ){
					gist.insert( fk, GappedIntervalSequenceTree.end );
					if( curry_len + fk.getLength() != gist.length() )
						System.err.println( "crap-o-la trois!" );
				}
				if( gk != null ){
					gist.insert( gk, GappedIntervalSequenceTree.end );
					if( curry_len + gk.getLength() != gist.length() )
						System.err.println( "crap-o-la cat!" );
				}
				fk = null;
				gk = null;

				// sequence ends here
				f_end_offset.addElement( new Long( cur_base + bufI ) );



				if( buf[ bufI ] == '>' ){
					section_start = bufI;
					state = read_defline;
					remaining_bytes++;
					bufI--;
				}
				else if( buf[ bufI ] == '=' ){
					// process a sequence set!
					
					if( seq_count == 0 )
						seq_count = lend.size();
					// create a new Match element
					Match m = new Match( seq_count );
					//copy file offsets 
					Match f_off_m = new Match( seq_count );
					// copy left ends into start
					m.starts = new long[ seq_count ];
					m.lengths = new long[ seq_count ];
					m.reverse = new boolean[ seq_count ];
					f_off_m.starts = new long[ seq_count ];
					f_off_m.lengths = new long[ seq_count ];
					int aligned_count = 0;
					for( int seqI = 0; seqI < seq_count; seqI++ ){
						m.starts[ seqI ] = ((Long)lend.elementAt( seqI )).longValue();
						if( m.starts[ seqI ] != 0 )
							aligned_count++;
						m.lengths[ seqI ] = ((Long)rend.elementAt( seqI )).longValue();
						m.reverse[ seqI ] = ((Boolean)reverse.elementAt( seqI )).booleanValue();
						f_off_m.starts[ seqI ] = ((Long)f_offset.elementAt( seqI )).longValue();
						f_off_m.lengths[ seqI ] = ((Long)f_end_offset.elementAt( seqI )).longValue();
					}
					// if this interval contains alignment of more than one sequence then call it an LCB
					// there may be problems with the assumption that an interval will always contain
					// some aligned sequence if it has more than one set of genome coordinates defined,
					// but let's not worry about that now since mauveAligner won't have that problem
					if( aligned_count > 1 ){
						LCB lcb = new LCB();
						lcb.left_end = new long[ seq_count ];
						lcb.right_end = new long[ seq_count ];
						lcb.reverse = new boolean[ seq_count ];
						lcb.left_adjacency = new int[ seq_count ];
						lcb.right_adjacency = new int[ seq_count ];
						lcb.lcb_id = tmp_lcbs.size();
						System.arraycopy( m.starts, 0, lcb.left_end, 0, seq_count );
						System.arraycopy( m.lengths, 0, lcb.right_end, 0, seq_count );
						System.arraycopy( m.reverse, 0, lcb.reverse, 0, seq_count );
						// set weight to average lcb length for now...
						long len_sum = 0;
						for( int seqI = 0; seqI < seq_count; seqI++ ){
							len_sum += lcb.right_end[ seqI ] - lcb.left_end[ seqI ];
						}
						lcb.weight = len_sum / seq_count;
						lcb.keep = false;
						tmp_lcbs.addElement( lcb );
					}
					
					
					// add to tmp_ivs
					tmp_ivs.addElement( m );
					tmp_offs.addElement( f_off_m );

					// clear data structures for the next alignment interval
					lend = new Vector();
					rend = new Vector();
					reverse = new Vector();
					f_offset = new Vector();
					f_end_offset = new Vector();


					state = read_comment;
					section_start = bufI;
					remaining_bytes++;
					bufI--;
				}
				break;

			}
			
			

			bufI++;
			remaining_bytes--;
		}
		
		intervals = new Match[ tmp_ivs.size() ];
		intervals = (Match[])tmp_ivs.toArray( intervals );
		file_pos = new Match[ tmp_offs.size() ];
		file_pos = (Match[])tmp_offs.toArray( file_pos );
		lcb_list = new LCB[ tmp_lcbs.size() ];
		lcb_list = (LCB[])tmp_lcbs.toArray( lcb_list );
		seq_length = new long[ seq_count ];
		gis_tree = new GappedIntervalSequenceTree[ intervals.length ][ seq_count ];
		for( int seqI = 0; seqI < seq_count; seqI++ ){
			long seq_len = 0;
			int treeI = 0;
			for( int ivI = 0; ivI < intervals.length; ivI++ ){
				if( intervals[ ivI ].lengths[ seqI ] > seq_len )
					seq_len = intervals[ ivI ].lengths[ seqI ];
				gis_tree[ ivI ][ seqI ] = (GappedIntervalSequenceTree)gis_tree_tmp.elementAt( treeI * seq_count + seqI );
				treeI++;
			}
			seq_length[ seqI ] = seq_len;
		}
		
		names = new String[ tmp_names.size() ];
		names = (String[])tmp_names.toArray( names );

		comments = new String[ tmp_comments.size() ];
		comments = (String[])tmp_comments.toArray( comments );
		// now test that the GIST works
//		testGIST();
		
		// verify the XMFA
//		validateXMFA();
	}
	
	public String getComment( int lcbI ){
		return comments[ lcbI ];
	}
	
	public String getName( int seqI ){
		return names[ seqI ];
	}

	/**
	 * Read columns of the alignment
	 */
	public String[] getColumns( int lcbI, long columnI, long count ) throws IOException{
	
		Match iv = intervals[ lcbI ];
		String[] cols = new String[ seq_count ];

		for( int seqI = 0; seqI < cols.length; seqI++ ){
			cols[ seqI ] = new String();
			long seek_offset = file_pos[ lcbI ].starts[ seqI ] + columnI + ((int)(columnI / line_width)) * newline_size;
			xmfa_file.seek( seek_offset );
			String cur_line;
			long remaining_cols = count;
			while( (cur_line = xmfa_file.readLine()) != null ){
				if( remaining_cols < cur_line.length() ){
					cols[ seqI ] += cur_line.substring(0, (int)remaining_cols );
					break;
				}else{
					cols[ seqI ] += cur_line;
					remaining_cols -= cur_line.length();
				}
			}
			
		}
		return cols;
	}
	
	/**
	 * Extracts columns from the sequence alignment containing the specified range 
	 * of the specified sequence.  The returned alignment columns will contain
	 * gaps and any whitespace in the XMFA source (e.g. newlines)
	 * @param seqI	The index of the sequence of interest
	 * @param lend	The left end coordinate of the range to be extracted
	 * @param rend	The right end coordinate of the range to be extracted
	 * @return		A set of alignment columns stored as an array of byte arrays
	 *				indexed as [sequence][column]
	 */
	public Object[] getRange( int seqI, long lend, long rend ) throws IOException{
		long cur_offset = lend;
		Object[] cols = new Object[ seq_count ];
		int seqJ = 0;
		for( seqJ = 0; seqJ < seq_count; seqJ++ )
			cols[ seqJ ] = new byte[0];

		while( cur_offset < rend ){
			// determine which LCB we start in
			int ivI = getLCB( seqI, cur_offset );
			
			// determine the file offset of the left end of this sequence
			// read a bunch o' columns
			long cur_iv_lend = intervals[ ivI ].starts[ seqI ];
			long read_size = intervals[ ivI ].lengths[ seqI ] - rend > 0 ? rend - cur_offset + 1 : intervals[ ivI ].lengths[ seqI ] - cur_offset + 1;
			while( read_size > 0 ){
				// assume forward orientation
				// plan b:  get the range of columns that need to be read
				//          for each seq, read the cols directly into a byte buffer
				//          reverse the sequences if necessary
				long lcb_offset = cur_offset - cur_iv_lend;
				long lcb_right_offset = intervals[ ivI ].reverse[ seqI ] ? intervals[ ivI ].lengths[ seqI ] - intervals[ ivI ].starts[ seqI ] - lcb_offset + 1 : lcb_offset + read_size;
				lcb_offset = intervals[ ivI ].reverse[ seqI ] ? lcb_right_offset - read_size : lcb_offset;

				long fk_left_col = gis_tree[ ivI ][ seqI ].seqPosToColumn( lcb_offset );
				long fk_right_col = gis_tree[ ivI ][ seqI ].seqPosToColumn( lcb_right_offset );

				Object[] byte_bufs = new Object[ seq_count ];
				for( seqJ = 0; seqJ < seq_count; seqJ++ ){
					byte_bufs[ seqJ ] = readRawSequence( ivI, seqJ, fk_left_col, fk_right_col - fk_left_col );
				}

				// just assume that the correct number of columns was read
				cur_offset += read_size;
				read_size = 0;
				
				// reverse the columns if necessary
				if( intervals[ ivI ].reverse[ seqI ] ){
					for( seqJ = 0; seqJ < seq_count; seqJ++ ){
						reverse( (byte[])byte_bufs[ seqJ ] );
					}
				}
				// append the columns to cols
				for( seqJ = 0; seqJ < seq_count; seqJ++ ){
					byte[] tmp_buf = new byte[ ((byte[])cols[seqJ]).length + ((byte[])byte_bufs[ seqJ ]).length ];
					System.arraycopy( (byte[])cols[seqJ], 0, tmp_buf, 0, ((byte[])cols[seqJ]).length );
					System.arraycopy( (byte[])byte_bufs[seqJ], 0, tmp_buf, ((byte[])cols[seqJ]).length, ((byte[])byte_bufs[seqJ]).length );
					cols[ seqJ ] = tmp_buf;
				}
			}
		}
		return cols;
	}
	
	/**
	 * reverse entries in a byte array
	 */
	void reverse( byte[] byte_buf ){
		for( int byteI = 0; byteI < byte_buf.length / 2; byteI++ ){
			byte tmp = byte_buf[ byteI ];
			byte_buf[ byteI ] = byte_buf[ byte_buf.length - byteI - 1 ];
			byte_buf[ byte_buf.length - byteI - 1 ] = tmp;
		}
	}
	
	/**
	 * filter gaps from one sequence while maintaining the columns of the alignment
	 * return the number of columns remaining after filtering
	 */
	int filterGapsInOneSequence( int seqI, Object[] byte_bufs ){
		// filter gaps from a particular sequence while maintaining column integrity
		int col_offset = 0;
		for( int colI = 0; colI < ((byte[])byte_bufs[ seqI ]).length; colI++ ){
			if( ((byte[])byte_bufs[ seqI ])[ colI ] == '-' )
				continue;
			// copy the column to the current col_offset
			try{
			for( int seqJ = 0; seqJ < seq_count; seqJ++ )
				((byte[])byte_bufs[ seqJ ])[ col_offset ] = ((byte[])byte_bufs[ seqJ ])[ colI ];
			}catch(Exception ee ){
				ee.printStackTrace();
			}
			col_offset++;
		}
		return col_offset;
	}

	/**
	 * Read sequence data (and any gap characters) from a file, filtering newlines
	 * column index starts at 0!!
	 * @param	ivI		The interval to read
	 * @param	seqI	The sequence to read
	 * @param	left_col	Left column index (interval local coordinates)
	 * @param	length		Length to read in columns (includes gaps)
	 */
	byte[] readRawSequence( int ivI, int seqI, long left_col, long length ) throws IOException{
		// check boundary condition
		if( gis_tree[ ivI ][ seqI ].size() == 0 ){
			if( length == 0 )
				return new byte[0];
			else
				throw new ArrayIndexOutOfBoundsException();
		}

		GistIterator l_iter = gis_tree[ ivI ][ seqI ].find( left_col );
		GistIterator r_iter = gis_tree[ ivI ][ seqI ].find( left_col + length - 1 );
		FileKey l_fk, r_fk;
		GapKey l_gk = null, r_gk = null;
		long l_gaps = 0;

	    // if the requested region contains nothing but gap then just return a buffer of gaps
		long seq_off = gis_tree[ ivI ][ seqI ].getSequenceStart( l_iter );
	    if( gis_tree[ ivI ][ seqI ].sequenceLength() <= left_col && 
	    	l_iter.node.seq_length == 0  && seq_off <= left_col ){
	    	if( left_col + length - 1 > gis_tree[ ivI ][ seqI ].length() )
	    		throw new ArrayIndexOutOfBoundsException();
	    	byte[] bb = new byte[ (int)length ];
	    	for( int bbI = 0; bbI < bb.length; bbI++ )
		    	bb[ bbI ] = (byte)'-';
		    return bb;
	    }


		try{
			l_fk = (FileKey)l_iter.node.key;
		}catch( ClassCastException cca ){
			// if we started in a gap the next one should be a FileKey
			try{
				l_gk = (GapKey)l_iter.node.key;
				seq_off = gis_tree[ ivI ][ seqI ].getSequenceStart( l_iter );
				l_iter = gis_tree[ ivI ][ seqI ].find_seqindex( seq_off );
				l_fk = (FileKey)l_iter.node.key;
			}catch( ClassCastException ccb ){
				System.err.println( "Something's messed up.  This shouldn't be a gap\n" );
				throw new IOException();
			}
		}
		try{
			if( r_iter.node == null )
				System.err.println("Error! fasfd report to master immediately");
			r_fk = (FileKey)r_iter.node.key;
		}catch( ClassCastException ccc ){
			// if we started in a gap the next one should be a FileKey
			try{
				r_gk = (GapKey)r_iter.node.key;
				seq_off = gis_tree[ ivI ][ seqI ].getSequenceStart( r_iter ) - 1;
				r_iter = gis_tree[ ivI ][ seqI ].find_seqindex( seq_off );
				r_fk = (FileKey)r_iter.node.key;
			}catch( ClassCastException ccd ){
				System.err.println( "Something's messed up.  This shouldn't be a gap\n" );
				throw new IOException();
			}
		}
		
		// should have l_fk and r_fk now.
		// get the file offsets to read from these
		long l_off = left_col - gis_tree[ ivI ][ seqI ].getStart( l_iter );
		// calculate the number of newlines in the space between the first desired
		// column and the first character in this key...
		long key_col_pos = gis_tree[ ivI ][ seqI ].getStart( l_iter ) % line_width;
		long l_newlines = ((key_col_pos + l_off) / 80) * newline_size;
		if( l_off < 0 ){
			// this happens when the desired column is a gap.  just skip ahead 
			l_gaps = -l_off;	// track how many gaps we'll need later
			l_newlines = 0;
			l_off = 0;
		}
		l_off += l_fk.f_offset + l_newlines;
		
		long r_off = left_col + length - 1 - gis_tree[ ivI ][ seqI ].getStart( r_iter );
		long r_key_col = gis_tree[ ivI ][ seqI ].getStart( r_iter ) % line_width;
		long r_newlines = ((r_key_col + r_off) / line_width) * newline_size;
		r_off += r_fk.f_offset + r_newlines;
		
		// now that the exact file offsets of the desired sequence have been
		// calculated, read it from the XMFA.
		if( r_off - l_off + 1 + l_gaps < 0 ){
			System.err.println("");
		}
		byte[] byte_buf = new byte[ (int)(r_off - l_off + 1 + l_gaps) ];
		for( int l_gapI = 0; l_gapI < l_gaps; l_gapI++ ){
			byte_buf[ l_gapI ] = (byte)'-';
		}
		xmfa_file.seek( l_off );
		xmfa_file.read( byte_buf, (int)l_gaps, byte_buf.length - (int)l_gaps );
		return byte_buf;
	}
	
	byte[] filterNewlines( byte[] byte_buf ){
		// filter the newlines
		int byte_off = 0;
        for( int byteI = 0; byteI < byte_buf.length; byteI++ ){
        	if( byte_buf[ byteI ] == '\r' || byte_buf[ byteI ] == '\n' )
        		continue;
        	byte_buf[ byte_off ] = byte_buf[ byteI ];
        	byte_off++;
        }
		byte[] bb2 = new byte[ byte_off ];
		System.arraycopy( byte_buf, 0, bb2, 0, byte_off );
		return bb2;
	}
	
	/**
	 * return the LCB index that contains the given position of the given sequence
	 */
	int getLCB( int sequence, long position ){
		int ivI = 0;
		for( ; ivI < intervals.length; ivI++ ){
			if( intervals[ ivI ].starts[ sequence ] <= position &&
				position <= intervals[ ivI ].lengths[ sequence ] )
				break;  // found the starting interval -- remember lengths array contains r_end
		}
		// throw an exception if the requested range couldn't be found
		if( ivI == intervals.length )
			throw new ArrayIndexOutOfBoundsException();
		return ivI;
	}

// FIXME:  get rev. comp right in these two functions!
	long revCompify( long position, int seq, int ivI ){
		return intervals[ ivI ].reverse[ seq ] ? intervals[ ivI ].lengths[ seq ] - intervals[ ivI ].starts[ seq ] - position : position;
	}
	// converts an LCB local coordinate to a global sequence coordinate, taking rev. comp. into account
	long globalToLCB( long position, int seq, int ivI ){
		long offset = position - intervals[ ivI ].starts[ seq ];
		return revCompify( offset, seq, ivI );
	}
	// converts a global sequence coordinate to an LCB local coordinate, taking rev. comp. into account
	long LCBToGlobal( long position, int seq, int ivI ){
		long offset = revCompify( position, seq, ivI );
		offset += intervals[ ivI ].starts[ seq ];
		if( intervals[ ivI ].starts[ seq ] == 0 )
			return 0;
		return offset;
	}
	
	class LcbStartComparator implements Comparator {
		int seqI;
		LcbStartComparator( int seqI ){
			this.seqI = seqI;
		}
		public int compare( Object o1, Object o2 ){
			LCB l1 = (LCB)o1;
			LCB l2 = (LCB)o2;
			return (int)(l1.left_end[ seqI ] - l2.left_end[ seqI ]);
		}
	}
	
	/**
	 * Checks the XMFA file to ensure that every nucleotide in the 
	 * relevant sequences are represented exactly once
	 * @return true if the parsed XMFA is valid
	 */
	boolean validateXMFA() throws IOException {
		boolean valid = true;
		LCB[] m_lcb_list = new LCB[ intervals.length ];
		for( int ivI = 0; ivI < intervals.length; ivI++ ){
			LCB lcb = new LCB();
			lcb.left_end = new long[ intervals[ ivI ].starts.length ];
			lcb.right_end = new long[ intervals[ ivI ].starts.length ];
			lcb.reverse = new boolean[ intervals[ ivI ].starts.length ];
			System.arraycopy( intervals[ ivI ].starts, 0, lcb.left_end, 0, lcb.left_end.length );
			System.arraycopy( intervals[ ivI ].lengths, 0, lcb.right_end, 0, lcb.right_end.length );
			m_lcb_list[ ivI ] = lcb;
		}

		for( int seqI = 0; seqI < seq_count; seqI++ ){
			// sort the lcb list on the current sequence
			LcbStartComparator lcb_comp = new LcbStartComparator( seqI );
			Arrays.sort( m_lcb_list, lcb_comp );
			long cur_coord = 0;
			for( int lcbI = 0; lcbI < m_lcb_list.length; lcbI++ ){
				if( m_lcb_list[ lcbI ].left_end[ seqI ] == 0 )
					continue;	// this LCB isn't defined for this seq, skip it
				if( cur_coord != m_lcb_list[ lcbI ].left_end[ seqI ] - 1 ){
					System.out.println( "Missing " + cur_coord + " from seq " + seqI );
					System.out.println( "m_lcb_list[ lcbI ].left_end[ seqI ] " + m_lcb_list[ lcbI ].left_end[ seqI ] );
					System.out.println();
					valid = false;
				}
				cur_coord = m_lcb_list[ lcbI ].right_end[ seqI ];
			}
		}
		
// special test case:
		int debug_val = 3277;
		long lento = gis_tree[ debug_val ][ 1 ].length();
		byte[] buff = (byte[])readRawSequence( debug_val, 1, 0, lento );


		//
		// read each lcb and verify that it has the correct number of characters
		//
		for( int ivI = 0; ivI < intervals.length; ivI++ ){
			for( int seqI = 0; seqI < seq_count; seqI++ ){
				if( ivI == debug_val )
					System.out.print("");
				long seq_len = intervals[ ivI ].lengths[ seqI ] - intervals[ ivI ].starts[ seqI ] + 1;
				if( intervals[ ivI ].lengths[ seqI ] == 0 && intervals[ ivI ].starts[ seqI ] == 0 )
					seq_len = 0;
				long len = gis_tree[ ivI ][ seqI ].length();
				if( seq_len != gis_tree[ ivI ][ seqI ].sequenceLength() ){
					System.out.println( "Error: conflicting interval and gis_tree sequence length records" );
					System.out.println( "At interval: " + ivI + " sequence: " + seqI + " iv seqlen: " + 
								seq_len + " gis_tree len: " + gis_tree[ ivI ][ seqI ].sequenceLength() );
					System.out.println( "left end: " + intervals[ ivI ].starts[ seqI ] + " right end: " +
								intervals[ ivI ].lengths[ seqI ] );
					valid = false;
				}
				byte[] buf = (byte[])readRawSequence( ivI, seqI, 0, len );
				long char_count = 0;
				for( int bufI = 0; bufI < buf.length; bufI++ )
					if( buf[ bufI ] != '\r' && buf[ bufI ] != '\n' && buf[ bufI ] != '-' )
						char_count++;
				if( char_count != seq_len ){
					System.out.println( "Error in interval " + ivI + " seq: " + seqI +
						"\nExpected " + seq_len + " chars but read " + char_count );
					System.out.println( "left end: " + intervals[ ivI ].starts[ seqI ] + " right end: " +
								intervals[ ivI ].lengths[ seqI ] );
					valid = false;
				}
			}
		}
		
		if( valid )
			System.out.println( "XMFA validated\n" );
		return valid;
	}
	
	

	/**
	 * Identifies the LCB and the column within the LCB of a given sequence position
	 */
	long[] getLCBAndColumn( int sequence, long position ) {
		// determine the LCB 
		int ivI = getLCB( sequence, position );
		
		long lcb_offset = globalToLCB( position, sequence, ivI );
		long fk_left_col = gis_tree[ ivI ][ sequence ].seqPosToColumn( lcb_offset );
		
		long[] lcb_and_col = new long[2];
		lcb_and_col[0] = ivI;
		lcb_and_col[1] = fk_left_col;
		return lcb_and_col;
	}
	
	/**
	 * Returns the sequence coordinates aligned in a given column
	 */
	long[] getColumnCoordinates( int lcb, long column ) {
		long[] seq_offsets = new long[ seq_count ];
		for( int seqI = 0; seqI < seq_count; seqI++ ){
			seq_offsets[ seqI ] = gis_tree[ lcb ][ seqI ].columnToSeqPos( column );
			seq_offsets[ seqI ] = LCBToGlobal( seq_offsets[ seqI ], seqI, lcb );
		}
		return seq_offsets;
	}
	
	/**
	 * Reorder the sequences to conform to the order given in new_order[]
	 */
	public void reorder( int new_order[] ){
		long[] tmp = new long[ new_order.length ];
		for( int seqI = 0; seqI < new_order.length; seqI++ )
			tmp[ seqI ] = seq_length[ new_order[ seqI ] ];
		System.arraycopy( tmp, 0, seq_length, 0, tmp.length );
		
		for( int ivI = 0; ivI < intervals.length; ivI++ )
			intervals[ ivI ].reorder( new_order );

		for( int fposI = 0; fposI < file_pos.length; fposI++ )
			file_pos[ fposI ].reorder( new_order );
		
		for( int lcbI = 0; lcbI < lcb_list.length; lcbI++ )
			lcb_list[ lcbI ].reorder( new_order );

		GappedIntervalSequenceTree[] tmp_tree = new GappedIntervalSequenceTree[ new_order.length ];
		for( int ivI = 0; ivI < intervals.length; ivI++ ){
			for( int seqI = 0; seqI < new_order.length; seqI++ )
				tmp_tree[ seqI ] = gis_tree[ ivI ][ new_order[ seqI ] ];
			System.arraycopy( tmp_tree, 0, gis_tree[ ivI ], 0, tmp_tree.length );
		}
	}

	void testGIST(){
		// test the GIST to ensure it maintains proper element
		// ordering through a series of queries
		java.util.Random randy = new java.util.Random();
		GistIterator iter = null;
		for( int searchI = 0; searchI < 10000; searchI++ ){
			long search_index = 0;
			if( randy.nextInt() % 2 == 1 ){
				search_index = randy.nextInt() % gis_tree[0][0].length();
				iter = gis_tree[0][0].find( search_index );
			}else{
				search_index = randy.nextInt() % gis_tree[0][0].sequenceLength();
				iter = gis_tree[0][0].find_seqindex( search_index );
			}
		}
		
		// now that we've searched the tree many times, write the 
		// sequence back out for comparison
		int node_count = (int)gis_tree[0][0].nodeCount();
		System.out.println("node count: " + node_count);
		System.out.println("seq length: " + gis_tree[0][0].sequenceLength() );
		System.out.println("total length: " + gis_tree[0][0].length() );
		int iters = 0;
		try{
		File my_outtie = new File( "gist_test.txt" );
		FileWriter out_writer = new FileWriter( my_outtie );
		iter = gis_tree[0][0].find(0);
		while( true ){
			FileKey fk = null;
			GapKey gk = null;
			iters++;
			try{
				fk = (FileKey)iter.node.key;
			}catch( ClassCastException cca ){
				// if we started in a gap the next one should be a FileKey
				try{
					gk = (GapKey)iter.node.key;
				}catch( ClassCastException ccb ){
					System.err.println( "Something's messed up.  This shouldn't be a gap\n" );
					throw new IOException();
				}
			}
			if( fk != null ){
				// write out the sequence
				byte[] raw_buf = new byte[ (int)fk.f_length ];
				xmfa_file.seek( fk.f_offset );
				xmfa_file.read( raw_buf, 0, (int)fk.f_length );
				int seqI = 0;
				// count the seq length minus newlines
				for( int rawI = 0; rawI < raw_buf.length; rawI++ )
					if( raw_buf[ rawI ] != '\r' && raw_buf[ rawI ] != '\n' )
						seqI++;
				// extract the sequence without newlines
				char[] seq_buf = new char[ seqI ];
				seqI = 0;
				for( int rawI = 0; rawI < raw_buf.length; rawI++ )
					if( raw_buf[ rawI ] != '\r' && raw_buf[ rawI ] != '\n' )
						seq_buf[ seqI++ ] = (char)raw_buf[ rawI ];
				// write the sequence without newlines
				out_writer.write( seq_buf );
			}
			if( gk != null ){
				// write out some gaps
				char[] gap_buf = new char[ (int)gk.getLength() ];
				for( int gapI = 0; gapI < gap_buf.length; gapI++ )
					gap_buf[ gapI ] = '-';
				out_writer.write( gap_buf );
			}

			if( !iter.hasNext())
				break;
			iter.next();
		}
		out_writer.flush();
		out_writer.close();
		}catch( Exception e ){
			e.printStackTrace();
		}
		System.out.println( "iters: " + iters );
		node_count = (int)gis_tree[0][0].nodeCount();
		System.out.println("node count: " + node_count);
		System.out.println("seq length: " + gis_tree[0][0].sequenceLength() );
		System.out.println("total length: " + gis_tree[0][0].length() );
	}
}

